import collections
import configparser
import importlib
import os
import pathlib
import subprocess
import tempfile
import time
import traceback
import types
import typing

import uqbar.apis


class Configuration:
    """
    Configuration.

    ..  container:: example

        Behavior at instantiation:

        * Looks for ``$HOME/.abjad/``.

        * Creates ``$HOME/.abjad/`` if directory does not exist.

        * Looks for ``$HOME/.abjad/abjad.cfg``.

        * Creates ``$HOME/.abjad/abjad.cfg`` if file does not exist.

        * Parses ``$HOME/.abjad/abjad.cfg``.

        * Provides default key-value pairs for pairs which fail validation.

        * Writes configuration changes to disk.

        * Creates Abjad output directory if directory does not exist.

    Supports mutable mapping dictionary interface.
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = "System configuration"

    __slots__ = ("_cached_configuration_directory", "_settings")

    _configuration_directory_name = ".abjad"

    _configuration_file_name = "abjad.cfg"

    # for caching
    _lilypond_version_string: typing.ClassVar[str | None] = None

    ### INITIALIZER ###

    def __init__(self):
        self._cached_configuration_directory = None
        if not os.path.exists(str(self.configuration_directory)):
            try:
                os.makedirs(str(self.configuration_directory))
            except (IOError, OSError):
                traceback.print_exc()
        old_contents = ""
        if self.configuration_file_path.exists():
            try:
                old_contents = self.configuration_file_path.read_text()
            except AttributeError:
                with self.configuration_file_path.open(mode="r") as f:
                    old_contents = f.read()
        configuration = self._configuration_from_string(old_contents)
        configuration = self._validate_configuration(configuration)
        new_contents = self._configuration_to_string(configuration)
        if not self._compare_configurations(old_contents, new_contents):
            try:
                with open(str(self.configuration_file_path), "w") as file_pointer:
                    print(new_contents, file=file_pointer)
            except (IOError, OSError):
                traceback.print_exc()
        self._settings = configuration
        self._make_missing_directories()

    ### SPECIAL METHODS ###

    def __delitem__(self, i) -> None:
        """
        Deletes item ``i`` from configuration.
        """
        del self._settings[i]

    def __eq__(self, argument):
        """
        Returns true when ``argument`` is configuratioin with same settings.
        """
        if isinstance(argument, type(self)):
            return self._settings == argument._settings
        return False

    def __getitem__(self, argument) -> typing.Any:
        """
        Gets item or slice identified by ``argument``.
        """
        return self._settings.__getitem__(argument)

    def __iter__(self) -> typing.Iterator[str]:
        """
        Iterates configuration settings.
        """
        for key in self._settings:
            yield key

    def __setitem__(self, i, argument) -> None:
        """
        Sets configuration item ``i`` to ``argument``.
        """
        self._settings[i] = argument

    ### PRIVATE METHODS ###

    def _compare_configurations(self, old, new):
        old = "\n".join(old.splitlines()[3:])
        new = "\n".join(new.splitlines()[3:])
        return old == new

    def _configuration_from_string(self, string):
        if "[main]" not in string:
            string = "[main]\n" + string
        config_parser = configparser.ConfigParser()
        try:
            config_parser.read_string(string)
            configuration = dict(config_parser["main"].items())
        except configparser.ParsingError:
            configuration = {}
        return configuration

    def _configuration_to_string(self, configuration):
        option_definitions = self._get_option_definitions()
        known_items, unknown_items = [], []
        for key, value in sorted(configuration.items()):
            if key in option_definitions:
                known_items.append((key, value))
            else:
                unknown_items.append((key, value))
        result = []
        for line in self._get_initial_comment():
            if line:
                result.append(f"# {line}")
            else:
                result.append("")
        for key, value in known_items:
            result.append("")
            if key in option_definitions:
                for line in option_definitions[key]["comment"]:
                    if line:
                        result.append(f"# {line}")
                    else:
                        result.append("")
            if value not in ("", None):
                result.append(f"{key!s} = {value!s}")
            else:
                result.append(f"{key!s} =")
        if unknown_items:
            result.append("")
            result.append("# User-specified keys:")
            for key, value in unknown_items:
                result.append("")
                if value not in ("", None):
                    result.append(f"{key!s} = {value!s}")
                else:
                    result.append(f"{key!s} =")
        string = "\n".join(result)
        return string

    def _get_config_specification(self):
        specs = self._get_option_specification()
        return [f"{key} = {value}" for key, value in sorted(specs.items())]

    def _get_current_time(self):
        return time.strftime("%d %B %Y %H:%M:%S")

    def _get_initial_comment(self):
        current_time = self._get_current_time()
        return [
            f"Abjad configuration file created on {current_time}.",
            "This file is interpreted by Python's ConfigParser ",
            "and follows ini syntax.",
        ]

    def _get_option_comments(self):
        options = self._get_option_definitions()
        comments = [(key, options[key]["comment"]) for key in options]
        return dict(comments)

    def _get_option_definitions(self):
        options = {
            "abjad_output_directory": {
                "comment": [
                    "Set to the directory where all Abjad-generated files",
                    "(such as PDFs and LilyPond files) should be saved.",
                    "Defaults to $HOME/.abjad/output/",
                ],
                "default": os.path.join(str(self.configuration_directory), "output"),
                "validator": str,
            },
            "lilypond_path": {
                "comment": [
                    "Lilypond executable path. Set to override dynamic lookup."
                ],
                "default": "lilypond",
                "validator": str,
            },
            "midi_player": {
                "comment": [
                    "MIDI player to open MIDI files.",
                    "When unset your OS should know how to open MIDI files.",
                ],
                "default": None,
                "validator": str,
            },
            "pdf_viewer": {
                "comment": [
                    "PDF viewer to open PDF files.",
                    "When unset your OS should know how to open PDFs.",
                ],
                "default": None,
                "validator": str,
            },
            "text_editor": {
                "comment": [
                    "Text editor to edit text files.",
                    "When unset your OS should know how to open text files.",
                ],
                "default": None,
                "validator": str,
            },
        }
        return options

    def _get_option_specification(self):
        options = self._get_option_definitions()
        specs = [(key, options[key]["spec"]) for key in options]
        return dict(specs)

    def _make_missing_directories(self):
        if not os.path.exists(self.abjad_output_directory):
            try:
                os.makedirs(self.abjad_output_directory)
            except (IOError, OSError):
                traceback.print_exc()

    def _validate_configuration(self, configuration):
        option_definitions = self._get_option_definitions()
        for key in option_definitions:
            if key not in configuration:
                configuration[key] = option_definitions[key]["default"]
            validator = option_definitions[key]["validator"]
            if isinstance(validator, type):
                if not isinstance(configuration[key], validator):
                    configuration[key] = option_definitions[key]["default"]
            else:
                if not validator(configuration[key]):
                    configuration[key] = option_definitions[key]["default"]
        for key in configuration:
            if configuration[key] in ("", "None"):
                configuration[key] = None
        return configuration

    ### PUBLIC PROPERTIES ###

    @property
    def abjad_directory(self) -> pathlib.Path:
        """
        Gets Abjad directory.
        """
        return pathlib.Path(__file__).parent.parent

    @property
    def abjad_output_directory(self) -> pathlib.Path:
        """
        Gets Abjad output directory.
        """
        if "abjad_output_directory" in self._settings:
            return pathlib.Path(self._settings["abjad_output_directory"])
        return self.configuration_directory / "output"

    @property
    def boilerplate_directory(self) -> pathlib.Path:
        """
        Gets Abjad boilerplate directory.
        """
        return self.abjad_directory.parent / "boilerplate"

    @property
    def configuration_directory(self) -> pathlib.Path:
        """
        Gets configuration directory.

        ..  container:: example

            >>> configuration = abjad.Configuration()
            >>> configuration.configuration_directory
            PosixPath('...')

        Defaults to $HOME/{directory_name}.

        Returns $TEMP/{directory_name} if $HOME is read-only or $HOME/{directory_name}
        is read-only.

        Also caches the initial result to reduce filesystem interaction.
        """
        if self._cached_configuration_directory is None:
            directory_name = self._configuration_directory_name
            home_directory = str(self.home_directory)
            flags = os.W_OK | os.X_OK
            if os.access(home_directory, flags):
                path = self.home_directory / directory_name
                if not path.exists() or (path.exists() and os.access(str(path), flags)):
                    self._cached_configuration_directory = path
                    return self._cached_configuration_directory
            path = pathlib.Path(tempfile.gettempdir()) / directory_name
            self._cached_configuration_directory = path
        return self._cached_configuration_directory

    @property
    def configuration_file_path(self) -> pathlib.Path:
        """
        Gets configuration file path.

        ..  container:: example

            >>> configuration = abjad.Configuration()
            >>> configuration.configuration_file_path
            PosixPath('...')

        """
        return self.configuration_directory / self._configuration_file_name

    @property
    def home_directory(self) -> pathlib.Path:
        """
        Gets home directory.

        ..  container:: example

            >>> configuration = abjad.Configuration()
            >>> configuration.home_directory
            PosixPath('...')

        """
        path = (
            os.environ.get("HOME")
            or os.environ.get("HOMEPATH")
            or os.environ.get("APPDATA")
            or tempfile.gettempdir()
        )
        return pathlib.Path(path).absolute()

    @property
    def lilypond_log_file_path(self) -> pathlib.Path:
        """
        Gets LilyPond log file path.
        """
        return self.abjad_output_directory / "lily.log"

    ### PUBLIC METHODS ###

    def get(self, *arguments, **keywords):
        """
        Gets a key.
        """
        return self._settings.get(*arguments, **keywords)

    def get_lilypond_version_string(self) -> str:
        """
        Gets LilyPond version string.

        ..  container:: example

            >>> configuration = abjad.Configuration()
            >>> configuration.get_lilypond_version_string() # doctest: +SKIP
            '2.19.84'

        """
        if self._lilypond_version_string is not None:
            return self._lilypond_version_string
        command = ["lilypond", "--version"]
        proc = subprocess.run(command, stdout=subprocess.PIPE)
        assert proc.stdout is not None
        lilypond_version_string = proc.stdout.decode().split()[2]
        Configuration._lilypond_version_string = lilypond_version_string
        return lilypond_version_string


def list_all_classes(modules="abjad", ignored_classes=None):
    """
    Lists all public classes defined in ``path``.

    ..  container:: example

        >>> for class_ in abjad.list_all_classes(modules="abjad"): class_
        <class 'abjad.bind.Wrapper'>
        <class 'abjad.configuration.Configuration'>
        <class 'abjad.contextmanagers.ContextManager'>
        <class 'abjad.contextmanagers.FilesystemState'>
        <class 'abjad.contextmanagers.ForbidUpdate'>
        <class 'abjad.contextmanagers.NullContextManager'>
        <class 'abjad.contextmanagers.ProgressIndicator'>
        <class 'abjad.contextmanagers.RedirectedStreams'>
        <class 'abjad.contextmanagers.TemporaryDirectory'>
        <class 'abjad.contextmanagers.TemporaryDirectoryChange'>
        <class 'abjad.contextmanagers.Timer'>
        <class 'abjad.contributions.ContributionsBySite'>
        <class 'abjad.cyclictuple.CyclicTuple'>
        <class 'abjad.duration.Duration'>
        <class 'abjad.duration.Offset'>
        <class 'abjad.exceptions.AssignabilityError'>
        <class 'abjad.exceptions.ImpreciseMetronomeMarkError'>
        <class 'abjad.exceptions.LilyPondParserError'>
        <class 'abjad.exceptions.MissingContextError'>
        <class 'abjad.exceptions.MissingMetronomeMarkError'>
        <class 'abjad.exceptions.ParentageError'>
        <class 'abjad.exceptions.PersistentIndicatorError'>
        <class 'abjad.exceptions.SchemeParserFinishedError'>
        <class 'abjad.exceptions.UnboundedTimeIntervalError'>
        <class 'abjad.exceptions.WellformednessError'>
        <class 'abjad.get.Lineage'>
        <class 'abjad.indicators.Arpeggio'>
        <class 'abjad.indicators.Articulation'>
        <class 'abjad.indicators.BarLine'>
        <class 'abjad.indicators.BeamCount'>
        <class 'abjad.indicators.BendAfter'>
        <class 'abjad.indicators.BreathMark'>
        <class 'abjad.indicators.Clef'>
        <class 'abjad.indicators.ColorFingering'>
        <class 'abjad.indicators.Dynamic'>
        <class 'abjad.indicators.Fermata'>
        <class 'abjad.indicators.Glissando'>
        <class 'abjad.indicators.InstrumentName'>
        <class 'abjad.indicators.KeyCluster'>
        <class 'abjad.indicators.KeySignature'>
        <class 'abjad.indicators.LaissezVibrer'>
        <class 'abjad.indicators.LilyPondLiteral'>
        <class 'abjad.indicators.Markup'>
        <class 'abjad.indicators.MetronomeMark'>
        <class 'abjad.indicators.Mode'>
        <class 'abjad.indicators.Ottava'>
        <class 'abjad.indicators.RehearsalMark'>
        <class 'abjad.indicators.Repeat'>
        <class 'abjad.indicators.RepeatTie'>
        <class 'abjad.indicators.ShortInstrumentName'>
        <class 'abjad.indicators.StaffChange'>
        <class 'abjad.indicators.StartBeam'>
        <class 'abjad.indicators.StartGroup'>
        <class 'abjad.indicators.StartHairpin'>
        <class 'abjad.indicators.StartPhrasingSlur'>
        <class 'abjad.indicators.StartPianoPedal'>
        <class 'abjad.indicators.StartSlur'>
        <class 'abjad.indicators.StartTextSpan'>
        <class 'abjad.indicators.StartTrillSpan'>
        <class 'abjad.indicators.StemTremolo'>
        <class 'abjad.indicators.StopBeam'>
        <class 'abjad.indicators.StopGroup'>
        <class 'abjad.indicators.StopHairpin'>
        <class 'abjad.indicators.StopPhrasingSlur'>
        <class 'abjad.indicators.StopPianoPedal'>
        <class 'abjad.indicators.StopSlur'>
        <class 'abjad.indicators.StopTextSpan'>
        <class 'abjad.indicators.StopTrillSpan'>
        <class 'abjad.indicators.TextMark'>
        <class 'abjad.indicators.Tie'>
        <class 'abjad.indicators.TimeSignature'>
        <class 'abjad.indicators.VoiceNumber'>
        <class 'abjad.instruments.Accordion'>
        <class 'abjad.instruments.AltoFlute'>
        <class 'abjad.instruments.AltoSaxophone'>
        <class 'abjad.instruments.AltoTrombone'>
        <class 'abjad.instruments.AltoVoice'>
        <class 'abjad.instruments.BaritoneSaxophone'>
        <class 'abjad.instruments.BaritoneVoice'>
        <class 'abjad.instruments.BassClarinet'>
        <class 'abjad.instruments.BassFlute'>
        <class 'abjad.instruments.BassSaxophone'>
        <class 'abjad.instruments.BassTrombone'>
        <class 'abjad.instruments.BassVoice'>
        <class 'abjad.instruments.Bassoon'>
        <class 'abjad.instruments.Cello'>
        <class 'abjad.instruments.ClarinetInA'>
        <class 'abjad.instruments.ClarinetInBFlat'>
        <class 'abjad.instruments.ClarinetInEFlat'>
        <class 'abjad.instruments.Contrabass'>
        <class 'abjad.instruments.ContrabassClarinet'>
        <class 'abjad.instruments.ContrabassFlute'>
        <class 'abjad.instruments.ContrabassSaxophone'>
        <class 'abjad.instruments.Contrabassoon'>
        <class 'abjad.instruments.EnglishHorn'>
        <class 'abjad.instruments.Flute'>
        <class 'abjad.instruments.FrenchHorn'>
        <class 'abjad.instruments.Glockenspiel'>
        <class 'abjad.instruments.Guitar'>
        <class 'abjad.instruments.Harp'>
        <class 'abjad.instruments.Harpsichord'>
        <class 'abjad.instruments.Instrument'>
        <class 'abjad.instruments.Marimba'>
        <class 'abjad.instruments.MezzoSopranoVoice'>
        <class 'abjad.instruments.Oboe'>
        <class 'abjad.instruments.Percussion'>
        <class 'abjad.instruments.Piano'>
        <class 'abjad.instruments.Piccolo'>
        <class 'abjad.instruments.SopraninoSaxophone'>
        <class 'abjad.instruments.SopranoSaxophone'>
        <class 'abjad.instruments.SopranoVoice'>
        <class 'abjad.instruments.StringNumber'>
        <class 'abjad.instruments.TenorSaxophone'>
        <class 'abjad.instruments.TenorTrombone'>
        <class 'abjad.instruments.TenorVoice'>
        <class 'abjad.instruments.Trumpet'>
        <class 'abjad.instruments.Tuba'>
        <class 'abjad.instruments.Tuning'>
        <class 'abjad.instruments.Vibraphone'>
        <class 'abjad.instruments.Viola'>
        <class 'abjad.instruments.Violin'>
        <class 'abjad.instruments.Xylophone'>
        <class 'abjad.label.ColorMap'>
        <class 'abjad.lilypondfile.Block'>
        <class 'abjad.lilypondfile.LilyPondFile'>
        <class 'abjad.lyproxy.LilyPondContext'>
        <class 'abjad.lyproxy.LilyPondEngraver'>
        <class 'abjad.lyproxy.LilyPondGrob'>
        <class 'abjad.lyproxy.LilyPondGrobInterface'>
        <class 'abjad.math.Infinity'>
        <class 'abjad.math.NegativeInfinity'>
        <class 'abjad.meter.Meter'>
        <class 'abjad.meter.MetricAccentKernel'>
        <class 'abjad.metricmodulation.MetricModulation'>
        <class 'abjad.obgc.OnBeatGraceContainer'>
        <class 'abjad.overrides.Interface'>
        <class 'abjad.overrides.LilyPondOverride'>
        <class 'abjad.overrides.LilyPondSetting'>
        <class 'abjad.overrides.OverrideInterface'>
        <class 'abjad.overrides.SettingInterface'>
        <class 'abjad.parentage.Parentage'>
        <class 'abjad.parsers.base.Parser'>
        <class 'abjad.pattern.Pattern'>
        <class 'abjad.pattern.PatternTuple'>
        <class 'abjad.pcollections.PitchClassSegment'>
        <class 'abjad.pcollections.PitchClassSet'>
        <class 'abjad.pcollections.PitchRange'>
        <class 'abjad.pcollections.PitchSegment'>
        <class 'abjad.pcollections.PitchSet'>
        <class 'abjad.pcollections.TwelveToneRow'>
        <class 'abjad.pitch.Accidental'>
        <class 'abjad.pitch.Interval'>
        <class 'abjad.pitch.IntervalClass'>
        <class 'abjad.pitch.NamedInterval'>
        <class 'abjad.pitch.NamedIntervalClass'>
        <class 'abjad.pitch.NamedInversionEquivalentIntervalClass'>
        <class 'abjad.pitch.NamedPitch'>
        <class 'abjad.pitch.NamedPitchClass'>
        <class 'abjad.pitch.NumberedInterval'>
        <class 'abjad.pitch.NumberedIntervalClass'>
        <class 'abjad.pitch.NumberedInversionEquivalentIntervalClass'>
        <class 'abjad.pitch.NumberedPitch'>
        <class 'abjad.pitch.NumberedPitchClass'>
        <class 'abjad.pitch.Octave'>
        <class 'abjad.pitch.Pitch'>
        <class 'abjad.pitch.PitchClass'>
        <class 'abjad.pitch.StaffPosition'>
        <class 'abjad.rhythmtrees.RhythmTreeContainer'>
        <class 'abjad.rhythmtrees.RhythmTreeLeaf'>
        <class 'abjad.rhythmtrees.RhythmTreeMixin'>
        <class 'abjad.rhythmtrees.RhythmTreeParser'>
        <class 'abjad.score.AfterGraceContainer'>
        <class 'abjad.score.BeforeGraceContainer'>
        <class 'abjad.score.Chord'>
        <class 'abjad.score.Cluster'>
        <class 'abjad.score.Component'>
        <class 'abjad.score.Container'>
        <class 'abjad.score.Context'>
        <class 'abjad.score.DrumNoteHead'>
        <class 'abjad.score.IndependentAfterGraceContainer'>
        <class 'abjad.score.Leaf'>
        <class 'abjad.score.MultimeasureRest'>
        <class 'abjad.score.Note'>
        <class 'abjad.score.NoteHead'>
        <class 'abjad.score.NoteHeadList'>
        <class 'abjad.score.Rest'>
        <class 'abjad.score.Score'>
        <class 'abjad.score.Skip'>
        <class 'abjad.score.Staff'>
        <class 'abjad.score.StaffGroup'>
        <class 'abjad.score.TremoloContainer'>
        <class 'abjad.score.Tuplet'>
        <class 'abjad.score.Voice'>
        <class 'abjad.select.LogicalTie'>
        <class 'abjad.setclass.SetClass'>
        <class 'abjad.tag.Tag'>
        <class 'abjad.timespan.OffsetCounter'>
        <class 'abjad.timespan.Timespan'>
        <class 'abjad.timespan.TimespanList'>
        <class 'abjad.tweaks.Bundle'>
        <class 'abjad.tweaks.Tweak'>
        <class 'abjad.verticalmoment.VerticalMoment'>

    """
    all_classes = set()
    for module in yield_all_modules(modules):
        if "parser" in module.__name__:
            continue
        name = module.__name__.split(".")[-1]
        for name in dir(module):
            item = getattr(module, name)
            if isinstance(item, type):
                if "sphinx" in repr(item):
                    continue
                if item.__name__.startswith("_"):
                    continue
                if "abjad.io" in str(item):
                    continue
                if "abjad" in repr(item):
                    all_classes.add(item)
    if ignored_classes:
        ignored_classes = set(ignored_classes)
        all_classes.difference_update(ignored_classes)
    return list(sorted(all_classes, key=lambda _: (_.__module__, _.__name__)))


def list_all_functions(modules="abjad"):
    """
    Lists all public functions defined in ``modules``.

    ..  container:: example

        >>> functions = abjad.list_all_functions(modules="abjad")
        >>> names = [_.__name__ for _ in functions]
        >>> # for name in sorted(names): name

    """
    all_functions = set()
    for module in yield_all_modules(modules):
        name = module.__name__.split(".")[-1]
        if name.startswith("_"):
            continue
        if "sphinx" in repr(module):
            continue
        for name in sorted(dir(module)):
            item = getattr(module, name)
            if isinstance(item, types.FunctionType):
                if item.__name__.startswith("_"):
                    continue
                if "abjad" not in repr(item.__module__):
                    continue
                all_functions.add(item)
    return list(sorted(all_functions, key=lambda _: (_.__module__, _.__name__)))


configuration = Configuration()


def yield_all_modules(paths=None):
    """
    Yields all modules encountered in ``path``.

    Returns generator.
    """
    _paths = []
    if not paths:
        _paths = configuration.abjad_directory
    elif isinstance(paths, str):
        module = importlib.import_module(paths)
        _paths.extend(module.__path__)
    elif isinstance(paths, types.ModuleType):
        _paths.extend(paths.__path__)
    elif isinstance(paths, collections.abc.Iterable):
        for path in paths:
            if isinstance(path, types.ModuleType):
                _paths.extend(path.__path__)
            elif isinstance(path, str):
                module = importlib.import_module(path)
                _paths.extend(module.__path__)
            else:
                raise ValueError(module)
    for path in _paths:
        for source_path in uqbar.apis.collect_source_paths([path]):
            package_path = uqbar.apis.source_path_to_package_path(source_path)
            yield importlib.import_module(package_path)



import dataclasses
import enum

from . import tag as _tag


class Sites(enum.Enum):
    """
    Contribution sites.
    """

    ABSOLUTE_BEFORE = enum.auto()
    BEFORE = enum.auto()
    OPEN_BRACKETS = enum.auto()
    OPENING = enum.auto()
    CONTENTS = enum.auto()
    CLOSING = enum.auto()
    CLOSE_BRACKETS = enum.auto()
    AFTER = enum.auto()
    ABSOLUTE_AFTER = enum.auto()


class Types(enum.Enum):
    """
    Contribution types.
    """

    ARTICULATIONS = enum.auto()
    COMMANDS = enum.auto()
    CONTEXT_SETTINGS = enum.auto()
    GROB_OVERRIDES = enum.auto()
    GROB_REVERTS = enum.auto()
    LEAK = enum.auto()
    LEAKS = enum.auto()
    MARKUP = enum.auto()
    PITCHED_TRILL = enum.auto()
    SPANNER_STARTS = enum.auto()
    SPANNER_STOPS = enum.auto()
    START_BEAM = enum.auto()
    STEM_TREMOLOS = enum.auto()
    STOP_BEAM = enum.auto()
    TRILL_SPANNER_STARTS = enum.auto()
    VOICE_NUMBER = enum.auto()


@dataclasses.dataclass(slots=True)
class _ContributionsByType:
    articulations: list[str] = dataclasses.field(default_factory=list)
    commands: list[str] = dataclasses.field(default_factory=list)
    leak: list[str] = dataclasses.field(default_factory=list)
    leaks: list[str] = dataclasses.field(default_factory=list)
    markup: list[str] = dataclasses.field(default_factory=list)
    pitched_trill: list[str] = dataclasses.field(default_factory=list)
    spanner_starts: list[str] = dataclasses.field(default_factory=list)
    spanner_stops: list[str] = dataclasses.field(default_factory=list)
    start_beam: list[str] = dataclasses.field(default_factory=list)
    stem_tremolos: list[str] = dataclasses.field(default_factory=list)
    stop_beam: list[str] = dataclasses.field(default_factory=list)
    trill_spanner_starts: list[str] = dataclasses.field(default_factory=list)
    voice_number: list[str] = dataclasses.field(default_factory=list)

    def __iter__(self):
        for type_ in (
            self.articulations,
            self.commands,
            self.leak,
            self.leaks,
            self.markup,
            self.pitched_trill,
            self.spanner_starts,
            self.spanner_stops,
            self.start_beam,
            self.stem_tremolos,
            self.stop_beam,
            self.trill_spanner_starts,
            self.voice_number,
        ):
            yield type_

    def tag(self, tag, deactivate=None):
        """
        Tags contributions.
        """
        self.articulations = _tag.double_tag(self.articulations, tag, deactivate)
        self.commands = _tag.double_tag(self.commands, tag, deactivate)
        self.leak = _tag.double_tag(self.leak, tag, deactivate)
        self.leaks = _tag.double_tag(self.leaks, tag, deactivate)
        self.markup = _tag.double_tag(self.markup, tag, deactivate)
        self.pitched_trill = _tag.double_tag(self.pitched_trill, tag, deactivate)
        self.spanner_starts = _tag.double_tag(self.spanner_starts, tag, deactivate)
        self.spanner_stops = _tag.double_tag(self.spanner_stops, tag, deactivate)
        self.start_beam = _tag.double_tag(self.start_beam, tag, deactivate)
        self.stem_tremolos = _tag.double_tag(self.stem_tremolos, tag, deactivate)
        self.stop_beam = _tag.double_tag(self.stop_beam, tag, deactivate)
        self.trill_spanner_starts = _tag.double_tag(
            self.trill_spanner_starts, tag, deactivate
        )
        self.voice_number = _tag.double_tag(self.voice_number, tag, deactivate)

    def update(self, contributions):
        """
        Updates contributions with ``contributions``.
        """
        assert isinstance(contributions, type(self))
        if contributions.articulations:
            self.articulations.append(contributions.articulations)
        if contributions.commands:
            self.commands.append(contributions.commands)
        if contributions.leak:
            self.leak.append(contributions.leak)
        if contributions.leaks:
            self.leaks.append(contributions.leaks)
        if contributions.markup:
            self.markup.append(contributions.markup)
        if contributions.pitched_trill:
            self.pitched_trill.append(contributions.pitched_trill)
        if contributions.spanner_starts:
            self.spanner_starts.append(contributions.spanner_starts)
        if contributions.spanner_stops:
            self.spanner_stops.append(contributions.spanner_stops)
        if contributions.start_beam:
            self.start_beam.append(contributions.start_beam)
        if contributions.stem_tremolos:
            self.stem_tremolos.append(contributions.stem_tremolos)
        if contributions.stop_beam:
            self.stop_beam.append(contributions.stop_beam)
        if contributions.trill_spanner_starts:
            self.trill_spanner_starts.append(contributions.trill_spanner_starts)
        if contributions.voice_number:
            self.voice_number.append(contributions.voice_number)


@dataclasses.dataclass(slots=True)
class ContributionsBySite:
    """
    LilyPond format contributions.
    """

    absolute_after: _ContributionsByType = dataclasses.field(
        default_factory=_ContributionsByType
    )
    absolute_before: _ContributionsByType = dataclasses.field(
        default_factory=_ContributionsByType
    )
    before: _ContributionsByType = dataclasses.field(
        default_factory=_ContributionsByType
    )
    after: _ContributionsByType = dataclasses.field(
        default_factory=_ContributionsByType
    )
    opening: _ContributionsByType = dataclasses.field(
        default_factory=_ContributionsByType
    )
    closing: _ContributionsByType = dataclasses.field(
        default_factory=_ContributionsByType
    )
    context_settings: list = dataclasses.field(default_factory=list)
    grob_overrides: list = dataclasses.field(default_factory=list)
    grob_reverts: list = dataclasses.field(default_factory=list)

    def __iter__(self):
        for site in (
            self.absolute_before,
            self.absolute_after,
            self.before,
            self.after,
            self.opening,
            self.closing,
            self.context_settings,
            self.grob_overrides,
            self.grob_reverts,
        ):
            yield site

    @staticmethod
    def alphabetize(lists):
        assert isinstance(lists, list)
        for item in lists:
            assert isinstance(item, list), repr(item)

        def key(list_):
            list_ = [_ for _ in list_ if not _.lstrip().startswith("%! ")]
            list_ = [_.removeprefix("%@% ") for _ in list_]
            return list_

        lists_ = lists[:]
        lists_.sort(key=key)
        strings = []
        for list_ in lists_:
            strings.extend(list_)
        return strings

    # TODO: rename or remove?
    def freeze_overrides(self):
        """
        Sorts each site.
        """
        self.context_settings = list(sorted(set(self.context_settings)))
        self.grob_overrides = list(sorted(set(self.grob_overrides)))
        self.grob_reverts = list(sorted(set(self.grob_reverts)))

    def get_contribution_lists(self):
        lists = []
        for site in self:
            for list_ in site:
                if list_:
                    lists.append(list_)
        return lists

    def tag_contributions(self, tag, deactivate=None):
        """
        Tags contributions with ``tag``.
        """
        self.absolute_before.tag(tag, deactivate)
        self.absolute_after.tag(tag, deactivate)
        self.before.tag(tag, deactivate)
        self.after.tag(tag, deactivate)
        self.opening.tag(tag, deactivate)
        self.closing.tag(tag, deactivate)
        self.context_settings = _tag.double_tag(self.context_settings, tag, deactivate)
        self.grob_overrides = _tag.double_tag(self.grob_overrides, tag, deactivate)
        self.grob_reverts = _tag.double_tag(self.grob_reverts, tag, deactivate)

    def update(self, contributions):
        """
        Updates contributions with ``contributions``.
        """
        assert isinstance(contributions, type(self)), repr(contributions)
        self.absolute_before.update(contributions.absolute_before)
        self.absolute_after.update(contributions.absolute_after)
        self.before.update(contributions.before)
        self.after.update(contributions.after)
        self.opening.update(contributions.opening)
        self.closing.update(contributions.closing)
        self.context_settings.extend(contributions.context_settings)
        self.grob_overrides.extend(contributions.grob_overrides)
        self.grob_reverts.extend(contributions.grob_reverts)



import collections

import abjad


def test_pcollections_sequence():
    """
    All pitch collections are registered as sequences.
    """

    for pcollection in (
        abjad.PitchClassSegment([0, 1, 2]),
        abjad.PitchClassSet([0, 1, 2]),
        abjad.PitchSegment([0, 1, 2]),
        abjad.PitchSet([0, 1, 2]),
        abjad.TwelveToneRow([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]),
    ):
        assert isinstance(pcollection, collections.abc.Sequence)



import copy

import pytest

import abjad


def test_Note___copy___01():
    """
    Copy note.
    """

    note_1 = abjad.Note(12, (1, 4))
    note_2 = copy.copy(note_1)

    assert isinstance(note_1, abjad.Note)
    assert isinstance(note_2, abjad.Note)
    assert abjad.lilypond(note_1) == abjad.lilypond(note_2)
    assert note_1 is not note_2


def test_Note___copy___02():
    """
    Copy note with LilyPond multiplier.
    """

    note_1 = abjad.Note("c''4", multiplier=(1, 2))
    note_2 = copy.copy(note_1)

    assert isinstance(note_1, abjad.Note)
    assert isinstance(note_2, abjad.Note)
    assert abjad.lilypond(note_1) == abjad.lilypond(note_2)
    assert note_1 is not note_2


def test_Note___copy___03():
    """
    Copy note with LilyPond grob overrides and LilyPond context settings.
    """

    note_1 = abjad.Note(12, (1, 4))
    abjad.override(note_1).Staff.NoteHead.color = "#red"
    abjad.override(note_1).Accidental.color = "#red"
    abjad.setting(note_1).tupletFullLength = True
    note_2 = copy.copy(note_1)

    assert isinstance(note_1, abjad.Note)
    assert isinstance(note_2, abjad.Note)
    assert abjad.lilypond(note_1) == abjad.lilypond(note_2)
    assert note_1 is not note_2


def test_Note___copy___04():
    """
    Copy note with grace container.
    """

    note_1 = abjad.Note("c'4")
    grace_container_1 = abjad.AfterGraceContainer([abjad.Note("d'32")])
    abjad.attach(grace_container_1, note_1)

    assert abjad.lilypond(note_1) == abjad.string.normalize(
        r"""
        \afterGrace
        c'4
        {
            d'32
        }
        """
    )

    note_2 = copy.copy(note_1)
    grace_container_2 = abjad.get.after_grace_container(note_2)

    assert abjad.lilypond(note_2) == abjad.string.normalize(
        r"""
        \afterGrace
        c'4
        {
            d'32
        }
        """
    )

    assert note_1 is not note_2
    assert grace_container_1 is not grace_container_2
    assert isinstance(grace_container_1, abjad.AfterGraceContainer)


def test_Note___copy___05():
    """
    Deepcopy orphan note.
    """

    note = abjad.Note("c'4")
    articulation = abjad.Articulation("staccato")
    abjad.attach(articulation, note)
    grace = abjad.BeforeGraceContainer("d'16")
    abjad.attach(grace, note)
    abjad.override(note).NoteHead.color = "#red"

    assert abjad.lilypond(note) == abjad.string.normalize(
        r"""
        \grace {
            d'16
        }
        \once \override NoteHead.color = #red
        c'4
        - \staccato
        """
    )

    new_note = copy.deepcopy(note)

    assert new_note is not note
    assert abjad.lilypond(new_note) == abjad.lilypond(note)


def test_Note___copy___06():
    """
    Deepcopy note in score.
    """

    staff = abjad.Staff("c'8 [ c'8 e'8 f'8 ]")
    note = staff[0]
    articulation = abjad.Articulation("staccato")
    abjad.attach(articulation, note)
    grace = abjad.BeforeGraceContainer("d'16")
    abjad.attach(grace, note)
    abjad.override(note).NoteHead.color = "#red"

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \grace {
                d'16
            }
            \once \override NoteHead.color = #red
            c'8
            - \staccato
            [
            c'8
            e'8
            f'8
            ]
        }
        """
    )

    new_note = copy.deepcopy(note)

    assert new_note is not note
    assert abjad.get.parentage(note).parent is staff
    assert abjad.get.parentage(new_note).parent is not staff
    assert isinstance(abjad.get.parentage(new_note).parent, abjad.Staff)
    assert abjad.lilypond(new_note) == abjad.lilypond(note)
    assert abjad.lilypond(abjad.get.parentage(note).parent) == abjad.lilypond(
        abjad.get.parentage(new_note).parent
    )


def test_Note___copy___07():
    """
    Copy note with tweaks in notehead.
    """

    note = abjad.Note("c'4")
    abjad.tweak(note.note_head, r"\tweak color #red")
    abjad.tweak(note.note_head, r"\tweak Accidental.color #red")
    copied_note = copy.copy(note)
    string = abjad.lilypond(copied_note)
    assert string == abjad.string.normalize(
        r"""
        \tweak Accidental.color #red
        \tweak color #red
        c'4
        """
    ), print(string)


def test_Note___init___01():
    """
    Initializes note from empty input.
    """

    note = abjad.Note()

    assert abjad.lilypond(note) == "c'4"


def test_Note___init___02():
    """
    Initializes note with pitch in octave zero.
    """

    note = abjad.Note(-37, (1, 4))

    assert abjad.lilypond(note) == "b,,,4"


def test_Note___init___03():
    """
    Initializes note with non-assignable duration.
    """

    with pytest.raises(abjad.AssignabilityError):
        abjad.Note(0, (5, 8))


def test_Note___init___04():
    """
    Initializes note with LilyPond-style pitch string.
    """

    note = abjad.Note("c,,", (1, 4))

    assert abjad.lilypond(note) == "c,,4"


def test_Note___init___05():
    """
    Initializes note with complete LilyPond-style note string.
    """

    note = abjad.Note("cs8.")

    assert abjad.lilypond(note) == "cs8."


def test_Note___init___06():
    """
    Initializes note from chord.
    """

    chord = abjad.Chord([2, 3, 4], (1, 4))
    note = abjad.Note(chord)

    assert abjad.lilypond(note) == abjad.string.normalize(
        r"""
        d'4
        """
    )

    assert abjad.wf.wellformed(note)


def test_Note___init___07():
    """
    Initializes note from tupletized chord.
    """

    chord = abjad.Chord([2, 3, 4], (1, 4))
    chords = abjad.mutate.copy(chord, 3)
    tuplet = abjad.Tuplet((2, 3), chords)
    note = abjad.Note(tuplet[0])

    assert abjad.lilypond(note) == abjad.string.normalize(
        r"""
        d'4
        """
    )

    assert abjad.wf.wellformed(note)


def test_Note___init___08():
    """
    Initializes note from beamed chord.
    """

    chord = abjad.Chord([2, 3, 4], (1, 8))
    chords = abjad.mutate.copy(chord, 3)
    voice = abjad.Voice(chords)
    abjad.beam(voice[:])
    note = abjad.Note(voice[0])

    assert abjad.lilypond(note) == abjad.string.normalize(
        r"""
        d'8
        [
        """
    ), print(abjad.lilypond(note))

    assert abjad.wf.wellformed(note)


def test_Note___init___09():
    """
    Initializes note from rest.
    """

    rest = abjad.Rest("r8")
    note = abjad.Note(rest)

    assert abjad.lilypond(note) == abjad.string.normalize(
        r"""
        8
        """
    )

    assert abjad.wf.wellformed(note)


def test_Note___init___10():
    """
    Initializes note from tupletized rest.
    """

    tuplet = abjad.Tuplet((2, 3), "r8 r8 r8")
    duration = tuplet[0].written_duration
    note = abjad.Note(tuplet[0])

    assert isinstance(tuplet[0], abjad.Rest)
    assert isinstance(note, abjad.Note)
    assert tuplet[0]._parent is tuplet
    assert tuplet[0].written_duration == duration
    assert note._parent is None


def test_Note___init___11():
    """
    Initializes note from beamed rest.
    """

    voice = abjad.Voice(
        [abjad.Note(0, (1, 8)), abjad.Rest((1, 8)), abjad.Note(0, (1, 8))]
    )
    abjad.beam(voice[:])
    note = abjad.Note(voice[1])

    assert isinstance(voice[1], abjad.Rest)
    assert isinstance(note, abjad.Note)
    assert voice[1]._parent is voice
    assert note._parent is None


def test_Note___init___12():
    """
    Initializes notes from skip.
    """

    skip = abjad.Skip((1, 8))
    duration = skip.written_duration
    note = abjad.Note(skip)

    assert isinstance(note, abjad.Note)
    assert dir(skip) == dir(abjad.Skip((1, 4)))
    assert dir(note) == dir(abjad.Note("c'4"))
    assert note._parent is None
    assert note.written_duration == duration


def test_Note___init___13():
    """
    Initializes note from tupletized skip.
    """

    tuplet = abjad.Tuplet((2, 3), "s8 s8 s8")
    duration = tuplet[0].written_duration
    note = abjad.Note(tuplet[0])

    assert isinstance(tuplet[0], abjad.Skip)
    assert isinstance(note, abjad.Note)
    assert tuplet[0]._parent is tuplet
    assert tuplet[0].written_duration == duration
    assert note._parent is None


def test_Note___init___14():
    """
    Initializes note from beamed skip.
    """

    voice = abjad.Voice(
        [abjad.Note(0, (1, 8)), abjad.Skip((1, 8)), abjad.Note(0, (1, 8))]
    )
    abjad.beam(voice[:])
    note = abjad.Note(voice[1])

    assert isinstance(voice[1], abjad.Skip)
    assert isinstance(note, abjad.Note)
    assert voice[1]._parent is voice
    assert note._parent is None


def test_Note___init___15():
    """
    Initializes note with cautionary accidental.
    """

    note = abjad.Note("c'?4")

    assert abjad.lilypond(note) == "c'?4"


def test_Note___init___16():
    """
    Initializes note with forced accidental.
    """

    note = abjad.Note("c'!4")

    assert abjad.lilypond(note) == "c'!4"


def test_Note___init___17():
    """
    Initializes note with both forced and cautionary accidental.
    """

    note = abjad.Note("c'!?4")

    assert abjad.lilypond(note) == "c'!?4"


def test_Note___init___18():
    """
    Initializes note from chord with forced and cautionary accidental.
    """

    chord = abjad.Chord("<c'!? e' g'>4")
    note = abjad.Note(chord)

    assert abjad.lilypond(note) == "c'!?4"


def test_Note___init___19():
    """
    Initialize note with drum pitch.
    """

    note = abjad.Note("sn4")

    assert abjad.lilypond(note) == "snare4"



import pytest

import abjad


def test_LilyPondParser__comments_01():
    target = abjad.Container([abjad.Note(0, (1, 4))])

    string = r"""
    { c'4 }
    % { d'4 }
    % { e'4 }"""  # NOTE: no newline should follow the final brace

    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result)


def test_LilyPondParser__containers__Container_01():
    parser = abjad.parser.LilyPondParser()
    target = abjad.Container()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__containers__Tuplet_01():
    notes = abjad.makers.make_notes([0, 2, 4], (1, 8))
    target = abjad.Tuplet((2, 3), notes)

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \tuplet 3/2
        {
            c'8
            d'8
            e'8
        }
        """
    )

    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__containers__nesting_01():
    target = abjad.Container(
        [abjad.Container([]), abjad.Container([abjad.Container([])])]
    )

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        {
            {
            }
            {
                {
                }
            }
        }
        """
    )

    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__containers__simultaneous_01():
    target = abjad.Container()
    target.simultaneous = True
    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__contexts__PianoStaff_01():
    target = abjad.StaffGroup(
        [
            abjad.Staff(abjad.makers.make_notes([0, 2, 4, 5, 7], (1, 8))),
            abjad.Staff(abjad.makers.make_notes([0, 2, 4, 5, 7], (1, 8))),
        ]
    )
    target.lilypond_type = "PianoStaff"

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new PianoStaff
        <<
            \new Staff
            {
                c'8
                d'8
                e'8
                f'8
                g'8
            }
            \new Staff
            {
                c'8
                d'8
                e'8
                f'8
                g'8
            }
        >>
        """
    )

    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__contexts__Score_01():
    target = abjad.Score()

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Score
        <<
        >>
        """
    )

    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__contexts__Staff_01():
    target = abjad.Staff([])

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Staff
        {
        }
        """
    )

    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__contexts__Staff_02():
    target = abjad.Staff([])
    target.simultaneous = True
    target.append(
        abjad.Voice(abjad.makers.make_notes([0, 2, 4, 5, 7, 9, 11, 12], (1, 8)))
    )
    target.append(
        abjad.Voice(abjad.makers.make_notes([0, 2, 4, 5, 7, 9, 11, 12], (1, 8)))
    )

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Staff
        <<
            \new Voice
            {
                c'8
                d'8
                e'8
                f'8
                g'8
                a'8
                b'8
                c''8
            }
            \new Voice
            {
                c'8
                d'8
                e'8
                f'8
                g'8
                a'8
                b'8
                c''8
            }
        >>
        """
    )

    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__contexts__StaffGroup_01():
    target = abjad.StaffGroup([])

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new StaffGroup
        <<
        >>
        """
    )

    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__contexts__Voice_01():
    target = abjad.Voice([])

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Voice
        {
        }
        """
    )

    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__contexts__context_ids_01():
    notes = abjad.makers.make_notes([0, 2, 4, 5, 7], (1, 8))
    target = abjad.Staff(notes)
    target.name = "foo"

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \context Staff = "foo"
        {
            c'8
            d'8
            e'8
            f'8
            g'8
        }
        """
    )

    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__contexts__with_blocks_01():
    target = abjad.Staff([])

    r"""
    \new Staff {
    }
    """

    string = r"""\new Staff \with { } {
    }
    """

    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__functions__grace_01():
    target = abjad.Container([abjad.Note("c'4"), abjad.Note("d'4"), abjad.Note("e'2")])

    grace = abjad.BeforeGraceContainer([abjad.Note("g''16"), abjad.Note("fs''16")])

    abjad.attach(grace, target[2])

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        {
            c'4
            d'4
            \grace {
                g''16
                fs''16
            }
            e'2
        }
        """
    )

    string = r"{ c'4 d'4 \grace { g''16 fs''16} e'2 }"
    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__functions__language_01():
    target = abjad.Container(
        [abjad.Note("cs'8"), abjad.Note("ds'8"), abjad.Note("ff'8")]
    )

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        {
            cs'8
            ds'8
            ff'8
        }
        """
    )

    string = r"\language nederlands { cis'8 dis'8 fes'8 }"
    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__functions__relative_01():
    pitches = [2, 5, 9, 7, 12, 11, 5, 2]
    notes = abjad.makers.make_notes(pitches, (1, 4))
    target = abjad.Container(notes)

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        {
            d'4
            f'4
            a'4
            g'4
            c''4
            b'4
            f'4
            d'4
        }
        """
    )

    string = r"\relative c' { d f a g c b f d }"
    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__functions__relative_02():
    pitches = [11, 12, 11, 14, 11, 16, 11, 9, 11, 7, 11, 5]
    notes = abjad.makers.make_notes(pitches, (1, 4))
    target = abjad.Container(notes)

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        {
            b'4
            c''4
            b'4
            d''4
            b'4
            e''4
            b'4
            a'4
            b'4
            g'4
            b'4
            f'4
        }
        """
    )

    string = r"\relative c'' { b c b d b e b a b g b f }"
    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__functions__relative_03():
    pitches = [9, -3, 12, 5, 7, 31, 9, 17]
    notes = abjad.makers.make_notes(pitches, (1, 4))
    target = abjad.Container(notes)

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        {
            a'4
            a4
            c''4
            f'4
            g'4
            g'''4
            a'4
            f''4
        }
        """
    )

    string = r"\relative c'' { a a, c' f, g g'' a,, f' }"
    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__functions__relative_04():
    pitches = [
        ["a'", "c''", "e''"],
        ["f'", "a'", "c''"],
        ["a'", "c''", "e''"],
        ["f''", "a''", "c'''"],
        ["b", "b'", "e''"],
    ]
    leaves = abjad.makers.make_leaves(pitches, 1)
    target = abjad.Container(leaves)

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        {
            <a' c'' e''>1
            <f' a' c''>1
            <a' c'' e''>1
            <f'' a'' c'''>1
            <b b' e''>1
        }
        """
    )

    string = r"""\relative c'' { <a c e>1 <f a c> <a c e> <f' a c> <b, e b,> }"""
    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__functions__relative_05():
    pitches = ["c", "f", "b", "e'", "a'", "d''", "g''", "c'''"]
    leaves = abjad.makers.make_leaves(pitches, [(1, 4)])
    target = abjad.Container(leaves)

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        {
            c4
            f4
            b4
            e'4
            a'4
            d''4
            g''4
            c'''4
        }
        """
    )

    string = r"""\relative c { c f b e a d g c }"""
    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__functions__relative_06():
    target = abjad.Container(
        [
            abjad.Note("c'", (1, 4)),
            abjad.Note("d'", (1, 4)),
            abjad.Note("e'", (1, 4)),
            abjad.Note("f'", (1, 4)),
            abjad.Container(
                [
                    abjad.Note("c''", (1, 4)),
                    abjad.Note("d''", (1, 4)),
                    abjad.Note("e''", (1, 4)),
                    abjad.Note("f''", (1, 4)),
                ]
            ),
        ]
    )

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        {
            c'4
            d'4
            e'4
            f'4
            {
                c''4
                d''4
                e''4
                f''4
            }
        }
        """
    )

    string = r"""\relative c' { c d e f \relative c'' { c d e f } }"""
    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__functions__relative_07():
    target = abjad.Container(
        [
            abjad.Note("d'", (1, 4)),
            abjad.Note("e'", (1, 4)),
            abjad.Container(
                [
                    abjad.Note("e", (1, 4)),
                    abjad.Note("fs", (1, 4)),
                    abjad.Container(
                        [abjad.Note("e'", (1, 4)), abjad.Note("fs'", (1, 4))]
                    ),
                ]
            ),
        ]
    )

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        {
            d'4
            e'4
            {
                e4
                fs4
                {
                    e'4
                    fs'4
                }
            }
        }
        """
    )

    string = r"""\relative c' { d e \transpose f g { d e \relative c' { d e } } }"""
    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__functions__relative_08():
    target = abjad.Container(
        [
            abjad.Note("c'", (1, 4)),
            abjad.Chord(["c'", "e'", "g'"], (1, 4)),
            abjad.Chord(["c''", "e''", "g'''"], (1, 4)),
            abjad.Chord(["e", "c'", "g''"], (1, 4)),
        ]
    )

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        {
            c'4
            <c' e' g'>4
            <c'' e'' g'''>4
            <e c' g''>4
        }
        """
    )

    string = r"""\relative c' { c <c e g> <c' e g'> <c, e, g''> }"""
    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__functions__relative_09():
    # http://lilypond.org/doc/v2.15/Documentation/c6/lily-8d84e2b9.ly
    pitches = ["c''", "fs''", "c''", "gf'", "b'", "ess''", "b'", "fff'"]
    notes = abjad.makers.make_notes(pitches, [(1, 2)])
    target = abjad.Container(notes)

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        {
            c''2
            fs''2
            c''2
            gf'2
            b'2
            ess''2
            b'2
            fff'2
        }
        """
    )

    string = r"""\relative c'' { c2 fs c2 gf b2 ess b2 fff }"""
    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__functions__relative_10():
    pitches = ["c''", "bf''", "cf'''", "c'''", "cf'''", "c''", "cf''", "c''"]
    notes = abjad.makers.make_notes(pitches, [(1, 4)])
    target = abjad.Container(notes)

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        {
            c''4
            bf''4
            cf'''4
            c'''4
            cf'''4
            c''4
            cf''4
            c''4
        }
        """
    )

    string = r"""\relative c'' { c bf' cf c cf c, cf c }"""
    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__functions__relative_11():
    pitches = ["cs''", "b''", "bs''", "cs'''", "bs''", "cs''", "bs'", "cs''"]
    notes = abjad.makers.make_notes(pitches, [(1, 4)])
    target = abjad.Container(notes)

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        {
            cs''4
            b''4
            bs''4
            cs'''4
            bs''4
            cs''4
            bs'4
            cs''4
        }
        """
    )

    string = r"""\relative cs'' { cs b' bs cs bs cs, bs cs }"""
    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__functions__transpose_01():
    pitches = ["e'", "gs'", "b'", "e''"]
    target = abjad.Staff(abjad.makers.make_notes(pitches, (1, 4)))
    key_signature = abjad.KeySignature(abjad.NamedPitchClass("e"), abjad.Mode("major"))
    abjad.attach(key_signature, target[0])

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \key e \major
            e'4
            gs'4
            b'4
            e''4
        }
        """
    )

    string = r"\transpose d e \relative c' \new Staff { \key d \major d4 fs a d }"
    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__functions__transpose_02():
    pitches = ["ef'", "f'", "g'", "bf'"]
    target = abjad.Staff(abjad.makers.make_notes(pitches, (1, 4)))
    key_signature = abjad.KeySignature(abjad.NamedPitchClass("ef"), abjad.Mode("major"))
    abjad.attach(key_signature, target[0])

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \key ef \major
            ef'4
            f'4
            g'4
            bf'4
        }
        """
    )

    string = r"\transpose a c' \relative c' \new Staff { \key c \major c4 d e g }"
    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__functions__transpose_03():
    target = abjad.Staff(
        [
            abjad.Container(
                abjad.makers.make_notes(["cs'", "ds'", "es'", "fs'"], (1, 4))
            ),
            abjad.Container(
                abjad.makers.make_notes(["df'", "ef'", "f'", "gf'"], (1, 4))
            ),
        ]
    )

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Staff
        {
            {
                cs'4
                ds'4
                es'4
                fs'4
            }
            {
                df'4
                ef'4
                f'4
                gf'4
            }
        }
        """
    )

    string = r"""music = \relative c' { c d e f }
    \new Staff {
        \transpose c cs \music
        \transpose c df \music
    }
    """

    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__indicators__Articulation_01():
    target = abjad.Staff(abjad.makers.make_notes(["c''"], [(1, 4)] * 6 + [(1, 2)]))
    articulation = abjad.Articulation("marcato")
    abjad.attach(articulation, target[0], direction=abjad.UP)
    articulation = abjad.Articulation("stopped")
    abjad.attach(articulation, target[1], direction=abjad.DOWN)
    articulation = abjad.Articulation("tenuto")
    abjad.attach(articulation, target[2])
    articulation = abjad.Articulation("staccatissimo")
    abjad.attach(articulation, target[3])
    articulation = abjad.Articulation("accent")
    abjad.attach(articulation, target[4])
    articulation = abjad.Articulation("staccato")
    abjad.attach(articulation, target[5])
    articulation = abjad.Articulation("portato")
    abjad.attach(articulation, target[6])

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Staff
        {
            c''4
            ^ \marcato
            c''4
            _ \stopped
            c''4
            - \tenuto
            c''4
            - \staccatissimo
            c''4
            - \accent
            c''4
            - \staccato
            c''2
            - \portato
        }
        """
    )

    string = r"""\new Staff { c''4^^ c''_+ c''-- c''-! c''4-> c''-. c''2-_ }"""

    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result
    for x in result:
        assert 1 == len(abjad.get.indicators(x, abjad.Articulation))


def test_LilyPondParser__indicators__Articulation_02():
    target = abjad.Staff([abjad.Note("c'", (1, 4))])
    articulation = abjad.Articulation("marcato")
    abjad.attach(articulation, target[0], direction=abjad.UP)
    articulation = abjad.Articulation("stopped")
    abjad.attach(articulation, target[0], direction=abjad.DOWN)
    articulation = abjad.Articulation("tenuto")
    abjad.attach(articulation, target[0])
    articulation = abjad.Articulation("staccatissimo")
    abjad.attach(articulation, target[0])
    articulation = abjad.Articulation("accent")
    abjad.attach(articulation, target[0])
    articulation = abjad.Articulation("staccato")
    abjad.attach(articulation, target[0])
    articulation = abjad.Articulation("portato")
    abjad.attach(articulation, target[0])

    string = r"""\new Staff { c'4 ^^ _+ -- -! -> -. -_ }"""

    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result
    assert 7 == len(abjad.get.indicators(result[0], abjad.Articulation))


def test_LilyPondParser__indicators__Articulation_03():
    target = abjad.Container(
        abjad.makers.make_notes(
            ["c''", "c''", "b'", "c''"], [(1, 4), (1, 4), (1, 2), (1, 1)]
        )
    )

    articulation = abjad.Articulation("staccato")
    abjad.attach(articulation, target[0])
    articulation = abjad.Articulation("mordent")
    abjad.attach(articulation, target[1])
    articulation = abjad.Articulation("turn")
    abjad.attach(articulation, target[2])
    articulation = abjad.Articulation("fermata")
    abjad.attach(articulation, target[3])

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        {
            c''4
            - \staccato
            c''4
            - \mordent
            b'2
            - \turn
            c''1
            - \fermata
        }
        """
    )

    string = r"""{ c''4\staccato c''\mordent b'2\turn c''1\fermata }"""

    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result
    for x in result:
        assert 1 == len(abjad.get.indicators(x, abjad.Articulation))


def test_LilyPondParser__indicators__BarLine_01():
    target = abjad.Staff(
        abjad.makers.make_notes(["e'", "d'", "c'"], [(1, 4), (1, 4), (1, 2)])
    )
    abjad.Score([target], name="Score")
    bar_line = abjad.BarLine("|.")
    abjad.attach(bar_line, target[-1])

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Staff
        {
            e'4
            d'4
            c'2
            \bar "|."
        }
        """
    )

    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result
    items = abjad.get.indicators(result[2])
    assert 1 == len(items) and isinstance(items[0], abjad.BarLine)


def test_LilyPondParser__indicators__Beam_01():
    target = abjad.Voice(abjad.makers.make_notes([0] * 4, [(1, 8)]))
    abjad.beam(target[0:3])
    abjad.beam(target[3:], beam_lone_notes=True)

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            c'8
            c'8
            ]
            c'8
            [
            ]
        }
        """
    )

    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__indicators__Beam_02():
    target = abjad.Voice(abjad.makers.make_notes([0] * 4, [(1, 8)]))
    abjad.beam(target[:])
    abjad.beam(target[1:3])

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            c'8
            [
            c'8
            ]
            c'8
            ]
        }
        """
    )

    with pytest.raises(Exception):
        abjad.LilyPondParser()(abjad.lilypond(target))


def test_LilyPondParser__indicators__Beam_03():
    target = abjad.Voice(abjad.makers.make_notes([0] * 4, [(1, 8)]))
    abjad.beam(target[:3])
    abjad.beam(target[2:])

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            c'8
            c'8
            [
            ]
            c'8
            ]
        }
        """
    )

    with pytest.raises(Exception):
        abjad.LilyPondParser()(abjad.lilypond(target))


def test_LilyPondParser__indicators__Beam_04():
    string = "{ c'8 [ c'8 c'8 c'8 }"
    with pytest.raises(Exception):
        abjad.LilyPondParser()(string)


def test_LilyPondParser__indicators__Beam_05():
    """
    With direction.
    """

    target = abjad.Voice(abjad.makers.make_notes(4 * [0], [(1, 8)]))
    start_beam = abjad.StartBeam()
    abjad.attach(start_beam, target[0], direction=abjad.UP)
    stop_beam = abjad.StopBeam()
    abjad.attach(stop_beam, target[2])
    start_beam = abjad.StartBeam()
    abjad.attach(start_beam, target[3], direction=abjad.DOWN)
    stop_beam = abjad.StopBeam()
    abjad.attach(stop_beam, target[3])

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            ^ [
            c'8
            c'8
            ]
            c'8
            _ [
            ]
        }
        """
    )

    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__indicators__Clef_01():
    target = abjad.Staff([abjad.Note(0, 1)])
    clef = abjad.Clef("bass")
    abjad.attach(clef, target[0])

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \clef "bass"
            c'1
        }
        """
    )

    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result
    clefs = abjad.get.indicators(result[0], abjad.Clef)
    assert len(clefs) == 1


def test_LilyPondParser__indicators__Dynamic_01():
    target = abjad.Voice("c2 c2 c2 c2 c2 c2")
    dynamic = abjad.Dynamic("ppp")
    abjad.attach(dynamic, target[0], direction=abjad.DOWN)
    dynamic = abjad.Dynamic("mp")
    abjad.attach(dynamic, target[1], direction=abjad.UP)
    dynamic = abjad.Dynamic("rfz")
    abjad.attach(dynamic, target[2])
    dynamic = abjad.Dynamic("mf")
    abjad.attach(dynamic, target[3])
    dynamic = abjad.Dynamic("spp")
    abjad.attach(dynamic, target[4])
    dynamic = abjad.Dynamic("ff")
    abjad.attach(dynamic, target[5])

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c2
            _ \ppp
            c2
            ^ \mp
            c2
            \rfz
            c2
            \mf
            c2
            \spp
            c2
            \ff
        }
        """
    )

    string = r"""\new Voice { c2 _ \ppp c ^ \mp c2\rfz c\mf c2\spp c\ff }"""
    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result)
    for leaf in result:
        dynamics = abjad.get.indicators(leaf, abjad.Dynamic)
        assert len(dynamics) == 1


def test_LilyPondParser__indicators__Glissando_01():
    target = abjad.Container([abjad.Note(0, 1), abjad.Note(0, 1)])
    abjad.glissando(target[:])
    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__indicators__Glissando_02():
    string = r"{ c \glissando }"
    with pytest.raises(Exception):
        abjad.LilyPondParser()(string)


def test_LilyPondParser__indicators__Glissando_03():
    string = r"{ \glissando c }"
    with pytest.raises(Exception):
        abjad.LilyPondParser()(string)


def test_LilyPondParser__indicators__Hairpin_01():
    target = abjad.Voice(abjad.makers.make_notes([0] * 5, [(1, 4)]))
    abjad.hairpin("< !", target[:3])
    abjad.hairpin("> ppp", target[2:])

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'4
            \<
            c'4
            c'4
            \!
            \>
            c'4
            c'4
            \ppp
        }
        """
    )

    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__indicators__Hairpin_02():
    """
    Dynamics can terminate hairpins.
    """

    target = abjad.Voice(abjad.makers.make_notes([0] * 3, [(1, 4)]))
    abjad.hairpin("<", target[0:2])
    abjad.hairpin("p > f", target[1:])

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'4
            \<
            c'4
            \p
            \>
            c'4
            \f
        }
        """
    )

    string = r"\new Voice \relative c' { c \< c \p \> c \f }"
    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__indicators__Hairpin_03():
    """
    Unterminated.
    """

    string = r"{ c \< c c c }"
    with pytest.raises(Exception):
        abjad.LilyPondParser()(string)


def test_LilyPondParser__indicators__Hairpin_04():
    """
    Unbegun is okay.
    """

    string = r"{ c c c c \! }"
    abjad.parser.LilyPondParser()(string)


def test_LilyPondParser__indicators__Hairpin_05():
    """
    No double dynamic spans permitted.
    """

    string = r"{ c \< \> c c c \! }"
    with pytest.raises(Exception):
        abjad.LilyPondParser()(string)


def test_LilyPondParser__indicators__Hairpin_06():
    """
    With direction.
    """

    target = abjad.Voice(abjad.makers.make_notes([0] * 5, [(1, 4)]))
    start_hairpin = abjad.StartHairpin("<")
    abjad.attach(start_hairpin, target[0], direction=abjad.UP)
    stop_hairpin = abjad.StopHairpin()
    abjad.attach(stop_hairpin, target[2])
    hairpin = abjad.StartHairpin(">")
    abjad.attach(hairpin, target[2], direction=abjad.DOWN)
    dynamic = abjad.Dynamic("ppp")
    abjad.attach(dynamic, target[-1])

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'4
            ^ \<
            c'4
            c'4
            \!
            _ \>
            c'4
            c'4
            \ppp
        }
        """
    )

    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__indicators__Hairpin_07():
    string = r"\new Staff { c'4 ( \p \< d'4 e'4 f'4 ) \! }"
    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(result) == abjad.string.normalize(
        r"""
        \new Staff
        {
            c'4
            \p
            (
            \<
            d'4
            e'4
            f'4
            )
            \!
        }
        """
    )


def test_LilyPondParser__indicators__HorizontalBracket_01():
    target = abjad.Voice(abjad.makers.make_notes([0] * 4, [(1, 4)]))
    abjad.horizontal_bracket(target[:])
    abjad.horizontal_bracket(target[:2])
    abjad.horizontal_bracket(target[2:])

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'4
            \startGroup
            \startGroup
            c'4
            \stopGroup
            c'4
            \startGroup
            c'4
            \stopGroup
            \stopGroup
        }
        """
    )

    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__indicators__HorizontalBracket_02():
    """
    Starting and stopping on the same leaf.
    """

    string = r"""{ c \startGroup \stopGroup c c c }"""
    with pytest.raises(Exception):
        abjad.LilyPondParser()(string)


def test_LilyPondParser__indicators__HorizontalBracket_03():
    """
    One group stopping on a leaf, while another begins on the same leaf.
    """

    string = r"""{ c \startGroup c \stopGroup \startGroup c c \stopGroup }"""
    with pytest.raises(Exception):
        abjad.LilyPondParser()(string)


def test_LilyPondParser__indicators__HorizontalBracket_04():
    """
    Unterminated.
    """

    string = r"""{ c \startGroup c c c }"""
    with pytest.raises(Exception):
        abjad.LilyPondParser()(string)


def test_LilyPondParser__indicators__HorizontalBracket_05():
    """
    Unstarted.
    """

    string = r"""{ c c c c \stopGroup }"""
    with pytest.raises(Exception):
        abjad.LilyPondParser()(string)


def test_LilyPondParser__indicators__KeySignature_01():
    target = abjad.Staff([abjad.Note("fs'", 1)])
    key_signature = abjad.KeySignature(abjad.NamedPitchClass("g"), abjad.Mode("major"))
    abjad.attach(key_signature, target[0])

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \key g \major
            fs'1
        }
        """
    )

    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result
    key_signatures = abjad.get.indicators(result[0], abjad.KeySignature)
    assert len(key_signatures) == 1


def test_LilyPondParser__indicators__Markup_01():
    target = abjad.Staff([abjad.Note(0, 1)])
    markup = abjad.Markup(r"\markup { hello! }")
    abjad.attach(markup, target[0], direction=abjad.UP)

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Staff
        {
            c'1
            ^ \markup { hello! }
        }
        """
    )

    string = r"""\new Staff { c'1 ^ "hello!" }"""

    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result
    assert 1 == len(abjad.get.markup(result[0]))


def test_LilyPondParser__indicators__Markup_02():
    target = abjad.Staff([abjad.Note(0, (1, 4))])
    markup = abjad.Markup(r'\markup { X Y Z "a b c" }')
    abjad.attach(markup, target[0], direction=abjad.DOWN)

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Staff
        {
            c'4
            _ \markup { X Y Z "a b c" }
        }
        """
    ), repr(target)


def test_LilyPondParser__indicators__Markup_03():
    """
    Articulations following markup block are (re)lexed correctly after
    returning to the "notes" lexical state after popping the "markup lexical state.
    """

    target = abjad.Staff([abjad.Note(0, (1, 4)), abjad.Note(2, (1, 4))])
    markup = abjad.Markup(r"\markup { hello }")
    abjad.attach(markup, target[0], direction=abjad.UP)
    articulation = abjad.Articulation(".")
    abjad.attach(articulation, target[0])

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Staff
        {
            c'4
            - \staccato
            ^ \markup { hello }
            d'4
        }
        """
    )

    string = r"""\new Staff { c' ^ \markup { hello } -. d' }"""

    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result
    assert 1 == len(abjad.get.markup(result[0]))


def test_LilyPondParser__indicators__Markup_04():
    string_1 = r"""\markup { \bold { A B C } \italic 123 }"""

    string_2 = r"""\markup { \bold
    {
        A
        B
        C
    } \italic
    123 }"""

    parser = abjad.parser.LilyPondParser()
    markup = parser(string_1)
    assert abjad.lilypond(markup) == string_2


def test_LilyPondParser__indicators__MetronomeMark_01():
    target = abjad.Score([abjad.Staff([abjad.Note(0, 1)])])
    mark = abjad.MetronomeMark(textual_indication='"As fast as possible"')
    abjad.attach(mark, target[0][0], context="Staff")

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Score
        <<
            \new Staff
            {
                \tempo "As fast as possible"
                c'1
            }
        >>
        """
    )

    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result
    leaves = abjad.select.leaves(result)
    leaf = leaves[0]
    marks = abjad.get.indicators(leaf, abjad.MetronomeMark)
    assert len(marks) == 1


def test_LilyPondParser__indicators__MetronomeMark_02():
    target = abjad.Score([abjad.Staff([abjad.Note(0, 1)])])
    leaves = abjad.select.leaves(target)
    mark = abjad.MetronomeMark(abjad.Duration(1, 4), 60)
    abjad.attach(mark, leaves[0], context="Staff")

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Score
        <<
            \new Staff
            {
                \tempo 4=60
                c'1
            }
        >>
        """
    )

    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result
    leaves = abjad.select.leaves(result)
    leaf = leaves[0]
    marks = abjad.get.indicators(leaf, abjad.MetronomeMark)
    assert len(marks) == 1


def test_LilyPondParser__indicators__MetronomeMark_03():
    target = abjad.Score([abjad.Staff([abjad.Note(0, 1)])])
    leaves = abjad.select.leaves(target)
    mark = abjad.MetronomeMark(abjad.Duration(1, 4), (59, 63))
    abjad.attach(mark, leaves[0], context="Staff")

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Score
        <<
            \new Staff
            {
                \tempo 4=59-63
                c'1
            }
        >>
        """
    )

    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result
    leaves = abjad.select.leaves(result)
    leaf = leaves[0]
    marks = abjad.get.indicators(leaf, abjad.MetronomeMark)
    assert len(marks) == 1


def test_LilyPondParser__indicators__MetronomeMark_04():
    target = abjad.Score([abjad.Staff([abjad.Note(0, 1)])])
    mark = abjad.MetronomeMark(
        reference_duration=abjad.Duration(1, 4),
        units_per_minute=60,
        textual_indication='"Like a majestic swan, alive with youth and vigour!"',
    )
    leaves = abjad.select.leaves(target)
    abjad.attach(mark, leaves[0], context="Staff")

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Score
        <<
            \new Staff
            {
                \tempo "Like a majestic swan, alive with youth and vigour!" 4=60
                c'1
            }
        >>
        """
    )

    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result
    leaves = abjad.select.leaves(result)
    leaf = leaves[0]
    marks = abjad.get.indicators(leaf, abjad.MetronomeMark)
    assert len(marks) == 1


def test_LilyPondParser__indicators__MetronomeMark_05():
    target = abjad.Score([abjad.Staff([abjad.Note(0, 1)])])
    mark = abjad.MetronomeMark(
        reference_duration=abjad.Duration(1, 16),
        units_per_minute=(34, 55),
        textual_indication='"Brighter than a thousand suns"',
    )
    leaves = abjad.select.leaves(target)
    abjad.attach(mark, leaves[0], context="Staff")

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Score
        <<
            \new Staff
            {
                \tempo "Brighter than a thousand suns" 16=34-55
                c'1
            }
        >>
        """
    )

    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result
    leaves = abjad.select.leaves(result)
    leaf = leaves[0]
    marksn = abjad.get.indicators(leaf, abjad.MetronomeMark)
    assert len(marksn) == 1


def test_LilyPondParser__indicators__PhrasingSlur_01():
    """
    Successful slurs, showing single leaf overlap.
    """

    target = abjad.Voice(abjad.makers.make_notes([0] * 4, [(1, 4)]))
    abjad.phrasing_slur(target[2:])
    abjad.phrasing_slur(target[:3])

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'4
            \(
            c'4
            c'4
            \)
            \(
            c'4
            \)
        }
        """
    )

    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__indicators__PhrasingSlur_02():
    """
    Swapped start and stop.
    """

    target = abjad.Voice(abjad.makers.make_notes([0] * 4, [(1, 4)]))
    abjad.phrasing_slur(target[2:])
    abjad.phrasing_slur(target[:3])

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'4
            \(
            c'4
            c'4
            \)
            \(
            c'4
            \)
        }
        """
    )

    string = r"\new Voice \relative c' { c \( c c \( \) c \) }"

    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__indicators__PhrasingSlur_03():
    """
    Single leaf.
    """

    string = r"{ c \( \) c c c }"
    with pytest.raises(Exception):
        abjad.LilyPondParser()(string)


def test_LilyPondParser__indicators__PhrasingSlur_04():
    """
    Unterminated.
    """

    string = r"{ c \( c c c }"
    with pytest.raises(Exception):
        abjad.LilyPondParser()(string)


def test_LilyPondParser__indicators__PhrasingSlur_05():
    """
    Unstarted.
    """

    string = r"{ c c c c \) }"
    with pytest.raises(Exception):
        abjad.LilyPondParser()(string)


def test_LilyPondParser__indicators__PhrasingSlur_06():
    """
    Nested.
    """

    string = r"{ c \( c \( c \) c \) }"
    with pytest.raises(Exception):
        abjad.LilyPondParser()(string)


def test_LilyPondParser__indicators__RepeatTie_01():
    target = abjad.Container([abjad.Note(0, 1), abjad.Note(0, 1)])
    repeat_tie = abjad.RepeatTie()
    abjad.attach(repeat_tie, target[-1])
    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__indicators__Slur_01():
    """
    Successful slurs, showing single leaf overlap.
    """

    target = abjad.Voice(abjad.makers.make_notes([0] * 4, [(1, 4)]))
    abjad.slur(target[2:])
    abjad.slur(target[:3])

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'4
            (
            c'4
            c'4
            )
            (
            c'4
            )
        }
        """
    )

    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__indicators__Slur_02():
    """
    Swapped start and stop.
    """

    target = abjad.Voice(abjad.makers.make_notes([0] * 4, [(1, 4)]))
    abjad.slur(target[2:])
    abjad.slur(target[:3])

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'4
            (
            c'4
            c'4
            )
            (
            c'4
            )
        }
        """
    )

    string = r"\new Voice \relative c' { c ( c c () c ) }"
    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__indicators__Slur_03():
    """
    Single leaf.
    """

    string = "{ c () c c c }"
    with pytest.raises(Exception):
        abjad.LilyPondParser()(string)


def test_LilyPondParser__indicators__Slur_04():
    """
    Unterminated.
    """

    string = "{ c ( c c c }"
    with pytest.raises(Exception):
        abjad.LilyPondParser()(string)


def test_LilyPondParser__indicators__Slur_05():
    """
    Unstarted.
    """

    string = "{ c c c c ) }"
    with pytest.raises(Exception):
        abjad.LilyPondParser()(string)


def test_LilyPondParser__indicators__Slur_06():
    """
    Nested.
    """

    string = "{ c ( c ( c ) c ) }"
    with pytest.raises(Exception):
        abjad.LilyPondParser()(string)


def test_LilyPondParser__indicators__Slur_07():
    """
    With direction.
    """

    target = abjad.Voice(abjad.makers.make_notes([0] * 4, [(1, 4)]))
    start_slur = abjad.StartSlur()
    abjad.slur(target[:3], direction=abjad.DOWN, start_slur=start_slur)
    start_slur = abjad.StartSlur()
    abjad.slur(target[2:], direction=abjad.UP, start_slur=start_slur)

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'4
            _ (
            c'4
            c'4
            )
            ^ (
            c'4
            )
        }
        """
    )

    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__indicators__StemTremolo_01():
    target = abjad.Staff([abjad.Note(0, 1)])
    stem_tremolo = abjad.StemTremolo(4)
    abjad.attach(stem_tremolo, target[0])

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Staff
        {
            c'1
            :4
        }
        """
    )

    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result
    stem_tremolos = abjad.get.indicators(result[0], abjad.StemTremolo)
    assert 1 == len(stem_tremolos)


def test_LilyPondParser__indicators__Text_01():
    """
    Successful text spanners, showing single leaf overlap.
    """

    container = abjad.Voice(abjad.makers.make_notes([0] * 4, [(1, 4)]))
    abjad.text_spanner(container[2:])
    abjad.text_spanner(container[:3])

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'4
            \startTextSpan
            c'4
            c'4
            \stopTextSpan
            \startTextSpan
            c'4
            \stopTextSpan
        }
        """
    )

    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(container))
    assert (
        abjad.lilypond(container) == abjad.lilypond(result) and container is not result
    )


def test_LilyPondParser__indicators__Text_02():
    """
    Swapped start and stop.
    """

    target = abjad.Voice(abjad.makers.make_notes([0] * 4, [(1, 4)]))
    abjad.text_spanner(target[2:])
    abjad.text_spanner(target[:3])

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'4
            \startTextSpan
            c'4
            c'4
            \stopTextSpan
            \startTextSpan
            c'4
            \stopTextSpan
        }
        """
    )

    string = (
        r"\new Voice \relative c' { c \startTextSpan c c \startTextSpan \stopTextSpan c"
        r" \stopTextSpan }"
    )
    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__indicators__Text_03():
    """
    Single leaf.
    """

    string = r"{ c \startTextSpan \stopTextSpan c c c }"
    with pytest.raises(Exception):
        abjad.LilyPondParser()(string)


def test_LilyPondParser__indicators__Text_04():
    """
    Unterminated.
    """

    string = r"{ c \startTextSpan c c c }"
    with pytest.raises(Exception):
        abjad.LilyPondParser()(string)


def test_LilyPondParser__indicators__Text_05():
    """
    Unstarted.
    """

    string = r"{ c c c c \stopTextSpan }"
    with pytest.raises(Exception):
        abjad.LilyPondParser()(string)


def test_LilyPondParser__indicators__Text_06():
    """
    Nested.
    """

    string = r"{ c \startTextSpan c \startTextSpan c \stopTextSpan c \stopTextSpan }"
    with pytest.raises(Exception):
        abjad.LilyPondParser()(string)


def test_LilyPondParser__indicators__Tie_01():
    target = abjad.Container([abjad.Note(0, 1), abjad.Note(0, 1)])
    abjad.tie(target[:])
    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__indicators__Tie_02():
    string = r"{ c ~ }"
    with pytest.raises(Exception):
        abjad.LilyPondParser()(string)


def test_LilyPondParser__indicators__Tie_03():
    string = r"{ ~ c }"
    with pytest.raises(Exception):
        abjad.LilyPondParser()(string)


def test_LilyPondParser__indicators__Tie_04():
    """
    With direction.
    """

    target = abjad.Container([abjad.Note(0, 1), abjad.Note(0, 1)])
    abjad.tie(target[:], direction=abjad.UP)
    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__indicators__Tie_05():
    """
    With direction.
    """

    target = abjad.Container([abjad.Note(0, 1), abjad.Note(0, 1)])
    abjad.tie(target[:], direction=abjad.DOWN)
    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__indicators__TimeSignature_01():
    target = abjad.Score([abjad.Staff([abjad.Note(0, 1)])])
    time_signature = abjad.TimeSignature((8, 8))
    abjad.attach(time_signature, target[0][0])

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Score
        <<
            \new Staff
            {
                \time 8/8
                c'1
            }
        >>
        """
    )

    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result
    leaves = abjad.select.leaves(result)
    leaf = leaves[0]
    time_signatures = abjad.get.indicators(leaf, abjad.TimeSignature)
    assert len(time_signatures) == 1


def test_LilyPondParser__indicators__Trill_01():
    """
    Successful trills, showing single leaf overlap.
    """

    notes = abjad.makers.make_notes(4 * [0], [(1, 4)])
    target = abjad.Voice(notes)
    abjad.trill_spanner(target[2:])
    abjad.trill_spanner(target[:3])

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'4
            \startTrillSpan
            c'4
            c'4
            \stopTrillSpan
            \startTrillSpan
            c'4
            \stopTrillSpan
        }
        """
    )

    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__indicators__Trill_02():
    """
    Swapped start and stop.
    """

    notes = abjad.makers.make_notes(4 * [0], [(1, 4)])
    target = abjad.Voice(notes)
    abjad.trill_spanner(target[2:])
    abjad.trill_spanner(target[:3])

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'4
            \startTrillSpan
            c'4
            c'4
            \stopTrillSpan
            \startTrillSpan
            c'4
            \stopTrillSpan
        }
        """
    )

    string = (
        r"\new Voice \relative c' { c \startTrillSpan c c \startTrillSpan \stopTrillSpan c"
        r" \stopTrillSpan }"
    )
    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__indicators__Trill_03():
    """
    Single leaf.
    """

    string = r"{ c \startTrillSpan \stopTrillSpan c c c }"
    with pytest.raises(Exception):
        abjad.LilyPondParser()(string)


def test_LilyPondParser__indicators__Trill_04():
    """
    Unterminated.
    """

    string = r"{ c \startTrillSpan c c c }"
    with pytest.raises(Exception):
        abjad.LilyPondParser()(string)


def test_LilyPondParser__indicators__Trill_05():
    """
    Unstarted.
    """

    string = r"{ c c c c \stopTrillSpan }"
    with pytest.raises(Exception):
        abjad.LilyPondParser()(string)


def test_LilyPondParser__indicators__Trill_06():
    """
    Nested.
    """

    string = (
        r"{ c \startTrillSpan c \startTrillSpan c \stopTrillSpan c \stopTrillSpan }"
    )
    with pytest.raises(Exception):
        abjad.LilyPondParser()(string)


def test_LilyPondParser__leaves__Chord_01():
    target = abjad.Chord([0, 1, 4], (1, 4))
    parser = abjad.parser.LilyPondParser()
    result = parser("{ %s }" % abjad.lilypond(target))
    assert (
        abjad.lilypond(target) == abjad.lilypond(result[0]) and target is not result[0]
    )


def test_LilyPondParser__leaves__MultiMeasureRest_01():
    target = abjad.MultimeasureRest((1, 4))
    parser = abjad.parser.LilyPondParser()
    result = parser("{ %s }" % abjad.lilypond(target))
    assert (
        abjad.lilypond(target) == abjad.lilypond(result[0]) and target is not result[0]
    )


def test_LilyPondParser__leaves__Note_01():
    target = abjad.Note(0, 1)
    parser = abjad.parser.LilyPondParser()
    result = parser("{ %s }" % abjad.lilypond(target))
    assert (
        abjad.lilypond(target) == abjad.lilypond(result[0]) and target is not result[0]
    )


def test_LilyPondParser__leaves__Rest_01():
    target = abjad.Rest((1, 8))
    parser = abjad.parser.LilyPondParser()
    result = parser("{ %s }" % abjad.lilypond(target))
    assert (
        abjad.lilypond(target) == abjad.lilypond(result[0]) and target is not result[0]
    )


def test_LilyPondParser__leaves__Skip_01():
    target = abjad.Skip((3, 8))
    parser = abjad.parser.LilyPondParser()
    result = parser("{ %s }" % abjad.lilypond(target))
    assert (
        abjad.lilypond(target) == abjad.lilypond(result[0]) and target is not result[0]
    )


def test_LilyPondParser__lilypondfile__LilyPondFile_01():
    string = "{ c } { c } { c } { c }"
    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert isinstance(result, abjad.LilyPondFile)


def test_LilyPondParser__lilypondfile__ScoreBlock_01():
    target = abjad.Block(name="score")
    target.items.append(abjad.Score())
    parser = abjad.parser.LilyPondParser()
    result = parser(abjad.lilypond(target))
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__misc__chord_repetition_01():
    target = abjad.Container(
        [
            abjad.Chord([0, 4, 7], (1, 4)),
            abjad.Chord([0, 4, 7], (1, 4)),
            abjad.Chord([0, 4, 7], (1, 4)),
            abjad.Chord([0, 4, 7], (1, 4)),
        ]
    )

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        {
            <c' e' g'>4
            <c' e' g'>4
            <c' e' g'>4
            <c' e' g'>4
        }
        """
    )

    string = r"""{ <c' e' g'> q q q }"""
    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__misc__chord_repetition_02():
    target = abjad.Voice(
        [
            abjad.Chord([0, 4, 7], (1, 8)),
            abjad.Chord([0, 4, 7], (1, 8)),
            abjad.Chord([0, 4, 7], (1, 4)),
            abjad.Chord([0, 4, 7], (3, 16)),
            abjad.Chord([0, 4, 7], (1, 16)),
            abjad.Chord([0, 4, 7], (1, 4)),
        ]
    )

    dynamic = abjad.Dynamic("p")
    abjad.attach(dynamic, target[0])
    articulation = abjad.Articulation("staccatissimo")
    abjad.attach(articulation, target[2])
    markup = abjad.Markup(r"\markup { text }")
    abjad.attach(markup, target[3], direction=abjad.UP)
    articulation = abjad.Articulation("staccatissimo")
    abjad.attach(articulation, target[-1])

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Voice
        {
            <c' e' g'>8
            \p
            <c' e' g'>8
            <c' e' g'>4
            - \staccatissimo
            <c' e' g'>8.
            ^ \markup { text }
            <c' e' g'>16
            <c' e' g'>4
            - \staccatissimo
        }
        """
    )

    string = r"""\new Voice { <c' e' g'>8\p q q4-! q8.^"text" q16 q4-! }"""
    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__misc__chord_repetition_03():
    target = abjad.Container(
        [
            abjad.Chord([0, 4, 7], (1, 8)),
            abjad.Note(12, (1, 8)),
            abjad.Chord([0, 4, 7], (1, 8)),
            abjad.Note(12, (1, 8)),
            abjad.Rest((1, 4)),
            abjad.Chord([0, 4, 7], (1, 4)),
        ]
    )

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        {
            <c' e' g'>8
            c''8
            <c' e' g'>8
            c''8
            r4
            <c' e' g'>4
        }
        """
    )

    string = r"""{ <c' e' g'>8 c'' q c'' r4 q }"""
    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__misc__comments_01():
    """
    Comments are ignored.
    """

    target = abjad.Staff()
    string = r"""\new Staff { %{ HOO HAH %} }"""
    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__misc__default_duration_01():
    target = abjad.Container(
        abjad.makers.make_notes(
            [0], [(1, 4), (1, 2), (1, 2), (1, 8), (1, 8), (3, 16), (3, 16)]
        )
    )
    target[-2].multiplier = (5, 17)
    target[-1].multiplier = (5, 17)

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        {
            c'4
            c'2
            c'2
            c'8
            c'8
            c'8. * 5/17
            c'8. * 5/17
        }
        """
    )

    string = r"""{ c' c'2 c' c'8 c' c'8. * 5/17 c' }"""

    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__misc__variables_01():
    target = abjad.Staff(
        [
            abjad.Container(
                [
                    abjad.Container(
                        [
                            abjad.Container(
                                [
                                    abjad.Container([abjad.Note(0, (1, 8))]),
                                    abjad.Note(2, (1, 8)),
                                    abjad.Note(4, (1, 4)),
                                ]
                            ),
                            abjad.Note(5, (1, 4)),
                            abjad.Note(7, (1, 2)),
                        ]
                    ),
                    abjad.Note(9, (1, 2)),
                    abjad.Note(11, 1),
                ]
            ),
            abjad.Note(12, 1),
        ]
    )

    assert abjad.lilypond(target) == abjad.string.normalize(
        r"""
        \new Staff
        {
            {
                {
                    {
                        {
                            c'8
                        }
                        d'8
                        e'4
                    }
                    f'4
                    g'2
                }
                a'2
                b'1
            }
            c''1
        }
        """
    )

    string = r"""
        foo = { c'8 }
        foo = { \foo d' e'4 }
        foo = { \foo f' g'2 }
        foo = { \foo a' b'1 }
        \new Staff { \foo c'' }
    """

    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser__misc__version_string_01():
    """
    Version strings are ignored.
    """

    target = abjad.Staff()
    string = r"""\version "2.14.2" \new Staff { }"""
    parser = abjad.parser.LilyPondParser()
    result = parser(string)
    assert abjad.lilypond(target) == abjad.lilypond(result) and target is not result


def test_LilyPondParser_accidentals_cautionary_01():
    string = "{ c?4 }"
    parsed = abjad.parser.LilyPondParser()(string)

    assert parsed[0].note_head.is_cautionary is True
    assert abjad.lilypond(parsed[0]) == "c?4"


def test_LilyPondParser_accidentals_cautionary_02():
    string = "{ <c? e g??>4 }"
    parsed = abjad.parser.LilyPondParser()(string)

    assert parsed[0].note_heads[0].is_cautionary is True
    assert parsed[0].note_heads[1].is_cautionary is False
    assert parsed[0].note_heads[2].is_cautionary is True
    assert abjad.lilypond(parsed[0]) == "<c? e g?>4"


def test_LilyPondParser_accidentals_forced_01():
    string = "{ c!4 }"
    parsed = abjad.parser.LilyPondParser()(string)

    assert parsed[0].note_head.is_forced is True
    assert abjad.lilypond(parsed[0]) == "c!4"


def test_LilyPondParser_accidentals_forced_02():
    string = "{ <c! e g!!>4 }"
    parsed = abjad.parser.LilyPondParser()(string)

    assert parsed[0].note_heads[0].is_forced is True
    assert parsed[0].note_heads[1].is_forced is False
    assert parsed[0].note_heads[2].is_forced is True
    assert abjad.lilypond(parsed[0]) == "<c! e g!>4"


def test_parse_rtm_syntax_01():
    rtm = "(3 (1 (3 (1 (3 (1 (3 (1 1 1 1))))))))"
    result = abjad.rhythmtrees.parse_rtm_syntax(rtm)

    assert abjad.lilypond(result) == abjad.string.normalize(
        r"""
        \tweak text #tuplet-number::calc-fraction-text
        \tuplet 4/3
        {
            c'4
            \tweak text #tuplet-number::calc-fraction-text
            \tuplet 4/3
            {
                c'4
                \tweak text #tuplet-number::calc-fraction-text
                \tuplet 4/3
                {
                    c'4
                    \tweak text #tuplet-number::calc-fraction-text
                    \tuplet 4/3
                    {
                        c'4
                        c'4
                        c'4
                        c'4
                    }
                }
            }
        }
        """
    )
